% A simple demonstration to illustrate the performance of the OS-CMA for iterative blind channel
% equalization and signal extraction.
%
% Similar experiments are reported in the references below:
%
% - V. Zarzoso, P. Comon, and D. Slock, <a href = "http://www.elsevier.com/books/handbook-of-blind-source-separation/comon/978-0-12-374726-6#description">"Semi-Blind Methods for Communications"</a>, 
%   in: P. Comon and C. Jutten (Eds.), Handbook of Blind Source Separation, Independent Component Analysis and
%   Applications, Oxford, UK: Academic Press, 2010, chapter 15, pp. 593-638.
%
% - V. Zarzoso et P. Comon, <a href = "http://www.lavoisier.fr/notice/fr2746215180.html">"Methodes semi-aveugles pour les telecommunications"</a>,
%   dans : C. Jutten et P. Comon, Separation de sources 2 : au-dela de l'aveugle et applications,
%   Paris : Hermes, 2007, chapitre 8, pp. 303-353.
%
% - V. Zarzoso and P. Comon, <a href = "http://www.i3s.unice.fr/~zarzoso/biblio/tsp05.pdf">"Blind and Semi-Blind Equalization Based on the Constant Power Criterion"</a>,
%   IEEE Transactions on Signal Processing, Vol. 53, No. 11, November 2005, pp. 4363-4375. 
%
%
% HISTORY:
%
%    <Please add modification date here>: - <please add modification details here>
%         
% -- 2014/11/21: Version 1 release ----------------------------------------------------------------
%
%    2010/02/18: - created by Vicente Zarzoso
%                  (I3S Laboratory, University of Nice Sophia Antipolis, CNRS, France).


%%% Some simulation parameters to play with

pole = 0.5;     % AR-channel pole location (keep modulus < 1 to guarantee stability)
T = 200;        % number of input symbols and observed signal length
SNR = 20;       % channel-output signal-to-noise ratio in dB         

L = 2;          % equalizer length; 
                % for optimal visualization, use L = 2 in this simulation (1-pole AR channel)

                
clc;
disp('=====================================================================');
disp('OPTIMAL STEP-SIZE CONSTANT MODULUS ALGORITHM (OS-CMA) - DEMO');
disp(' ');

%%% Generate source signal, noise and channel output
disp('>>> Generating source signal, noise and channel output...');
disp(' ');

s = (rand(1, T) > 0.5) + sqrt(-1)*(rand(1, T) > 0.5);   % QPSK modulation
s = s - mean(s);                % remove mean
s = s/sqrt(mean(abs(s).^2));    % unit-power normalization

S = toeplitz(s(L:-1:1).', s(L:T));  % source data matrix

gamma = 1; % constellation parameter for chosen source modulation

n = randn(1, T) + sqrt(-1)*randn(1, T); % complex AWGN
n = n - mean(n);
n = n/sqrt(mean(abs(n).^2));        

b = 1; a = [1, -pole];      % a simple 1-pole AR channel 
r = filter(b, a, s);        % noiseless channel output

Pr = mean(abs(r).^2);       % noiseless channel output power
Pn = Pr*10^(-SNR/10);       % noise power for given SNR level

n = sqrt(Pn)*n;

x = r + n;         % noisy channel output

disp(['1-AR channel pole location: z = ', num2str(pole)]);
disp(['observed signal length: T = ', num2str(T), ' symbols']);
SNR_dB = 10*log10(mean(abs(r).^2)/mean(abs(n).^2));
disp(['channel output SNR = ', num2str(SNR_dB), ' dB']);
disp(['equalizer length: L = ', num2str(L)]);

X = toeplitz(x(L:-1:1).', x(L:T));    % observed data matrix for a 2-tap equalizer

[channel_output_mse, delay] = min(mean((abs((S*X(L,:)')*X(L,:)/sum(abs(X(L,:)).^2) - S).^2)'));
channel_output_mse_dB = 10*log10(channel_output_mse)

wini = 4*rand(L, 1) - 2;    % random equalizer initialization, with coefficients between -2 and 2
tol  = 1e-3;        % termination test threshold
maxit = 100;        % maximum number of iterations

disp(' ');
disp('<PAUSE> Please press any key to continue');
disp(' '); disp(' ');
pause


%%% MMSE bound (assumes all symbols in the transmitted signal block are
%%% used for training -> performance bound for linear equalization)

disp('>>> MMSE bound (all symbols in transmitted frame used for training)'); disp(' ');

disp(['MMSE equalizers at delays 0 to ', num2str(L-1), ':']);

W_mmse = inv(X*X')*X*S'     % MMSE receivers at delays 0 (optimal) and 1 (suboptimal)

Y_mmse = W_mmse'*X;         % corresponding MMSE equalizer outputs

disp('MMSE equalizers'' output MSE (dB):');

mmse_mse = mean((abs(Y_mmse - S).^2)'); % MMSE equalizer MSE's
[mmse_mse_min, optimum_mmse_delay] = min(mmse_mse);
mmse_mse_dB = 10*log10(mmse_mse)

mmse_mse_improvement_dB = channel_output_mse_dB - mmse_mse_dB

optimum_mmse_delay = optimum_mmse_delay - 1 

disp(' ');
disp('<PAUSE> Please press any key to continue');
disp(' '); disp(' ');
pause


%%% OS-CMA algorithm
disp('>>> OS-CMA algorithm');
disp(' ');

[w, it, cmval, muopt, W] = os_cma(X, gamma, wini, tol, maxit);

cmval_dB = 10*log10(cmval);

disp('OS-CMA equalizer:');
w_oscma = w

number_of_iterations = it

y = w'*X;

[oscma_output_mse, delay] = min(mean((abs((S*y')*y/sum(abs(y).^2) - S).^2)'));    % compare using optimum scaling

oscma_output_mse_dB = 10*log10(oscma_output_mse)
equalization_delay = delay - 1

oscma_mse_improvement_dB = channel_output_mse_dB - oscma_output_mse_dB

% compute equalizer output MSE for each iteration
for k = 1:it+1
    y = W(:, k)'*X;
    oscma_mse(k) =  min(mean((abs((S*y')*y/sum(abs(y).^2) - S).^2)'));
end
oscma_mse_dB = 10*log10(oscma_mse);

disp(' ');
disp('<PAUSE> Please press any key to show OS-CMA equalizer evolution');
disp(' '); disp(' ');
pause

% obtain CM criterion surface
npoints = 41;
ww = linspace(-2, 2, npoints);
JJ = zeros(npoints, npoints);
[WW1, WW2] = meshgrid(ww, ww);
for i = 1:npoints
    for j = 1:npoints
        JJ(i, j) = 10*log10(cm_cost([WW1(i, j), WW2(i, j), zeros(1, L-2)].', X, gamma));
    end
end
cmval_dB_min = min([min(min(JJ)), min(cmval_dB(1:it+1))]);    % a reasonable approximation of the min CM cost 
                                                              % value achievable in the given scenario

fig_title = ['OS-CMA equalization: T = ', num2str(T), ' samples, SNR = ', num2str(SNR), ' dB'];
figure; set(gcf, 'Name', [fig_title, '; equalizer output before starting the OS-CMA iterations']);
set(gcf, 'Position', [200, 150, 1000, 800]);

% initialize subplots
subplot(2, 2, 1);   % CM criterion contour curves
contour(WW1, WW2, JJ, 30); hold on;
plot(real(W_mmse(1, 1)), real(W_mmse(2, 1)), 'or', 'LineWidth', 2); % MMSE equalizer at delay 0
plot(-real(W_mmse(1, 1)), -real(W_mmse(2, 1)), 'or', 'LineWidth', 2); % MMSE equalizer at delay 0 (up to scale)
plot(real(W_mmse(1, 2)), real(W_mmse(2, 2)), 'sr', 'LineWidth', 2); % MMSE equalizer at delay 1
plot(-real(W_mmse(1, 2)), -real(W_mmse(2, 2)), 'sr', 'LineWidth', 2); % MMSE equalizer at delay 1 (up to scale)
plot(real(W(1, 1)), real(W(2, 1)), '+');            % initial equalizer vector
set(gca, 'DataAspectRatio', [1, 1, 1]);
xlabel('w_1'); ylabel('w_2');
title('CM cost and OS-CMA equalizer trajectories')

subplot(2, 2, 2);   % CM function values
plot(-1, 1, '-'); hold on; plot(-1, 1, ':');   % dummy plots to get legend right
plot(0, cmval_dB(1), '+'); 
plot([0, it], cmval_dB_min*ones(1, 2), ':');
axis([0, it, cmval_dB_min - 5, max(cmval_dB)+5]);
xlabel('iteration number'); ylabel('10 log_{10}(J_{CM})');
title('Equalizer-output CM cost')
legend('OS-CMA', 'minimum CM cost', 1);

subplot(2, 2, 3);   % equalizer-output eye diagram
y = W(:, 1)'*X;
polar(angle(y), abs(y), 'o')
set(gca, 'DataAspectRatio', [1, 1, 1]); grid;
xlabel('Re(y)'); ylabel('Im(y)');
title('Equalizer-output eye diagram')

subplot(2, 2, 4);   % equalizer-output MSE values
plot(-1, 1, '-'); hold on; plot(-1, 1, ':');    % some dummy plots to get legend right
plot(0, oscma_mse_dB(1), '+');
plot([0, it], 10*log10(mmse_mse_min)*ones(1, 2), ':');
axis([0, it, 10*log10(mmse_mse_min)-5, 0]);
title('Equalizer-output MSE')
xlabel('iteration number'); ylabel('MSE (dB)');
legend('OS-CMA', 'optimum-delay MMSE', 1)

disp('>>> Equalizer output before starting the OS-CMA iterations');
disp(' ');
disp('o      : optimal-delay MMSE equalizers');
disp('square : suboptimal-delay MMSE equalizers');
disp('+      : initial OS-CMA equalizer');
disp('x      : final OS-CMA equalizer');

oscma_mse_init_dB = oscma_mse_dB(1)
oscma_mse_improvement_dB = channel_output_mse_dB - oscma_mse_dB(1)
disp(' ');
disp('<PAUSE> Please press any key to continue');
disp(' '); disp(' ');
pause

it_plot = 0;    % iteration count

% loop over equalizer output after each iteration
while it_plot < it

it_plot = it_plot + 1;

subplot(2, 2, 1);   % OS-CMA equalizer trajectories
plot(real(W(1, it_plot:it_plot+1)), real(W(2, it_plot:it_plot+1)), '-')

subplot(2, 2, 2);   % CM function values
plot(it_plot-1:it_plot, cmval_dB(it_plot:it_plot+1), '-')

subplot(2, 2, 3);   % equalizer-output eye diagram 
y = W(:, it_plot+1)'*X;
polar(angle(y), abs(y), 'o')
set(gca, 'DataAspectRatio', [1, 1, 1]); grid;
xlabel('Re(y)'); ylabel('Im(y)'); title('Equalizer-output eye diagram')

subplot(2, 2, 4);   % equalizer-output MSE values
plot(it_plot-1:it_plot, oscma_mse_dB(it_plot:it_plot+1), '-');
axis([0, it, 10*log10(mmse_mse_min)-5, 1]);

set(gcf, 'Name', [fig_title, '; OS-CMA equalizer output after iteration #', num2str(it_plot)]);

disp(['>>> OS-CMA equalizer output after iteration #', num2str(it_plot)]);
oscma_mse_it_dB = oscma_mse_dB(it_plot+1)
oscma_mse_improvement_dB = channel_output_mse_dB - oscma_mse_dB(it_plot+1)
disp(' ');
disp('<PAUSE> Please press any key to continue');
disp(' '); disp(' ');
pause

end % while it_plot

subplot(2, 2, 1);
plot(real(W(1, it_plot:it_plot+1)), real(W(2, it_plot:it_plot+1)), 'x', 'LineWidth', 2);    % final equalizer


disp('OPTIMAL STEP-SIZE CONSTANT MODULUS ALGORITHM (OS-CMA) - END OF DEMO');
disp('=====================================================================');


